"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.QuestDriver = void 0;
const pg_1 = require("pg");
const moment = __importStar(require("moment"));
const ramda_1 = __importDefault(require("ramda"));
const query_orchestrator_1 = require("@cubejs-backend/query-orchestrator");
const QuestQuery_1 = require("./QuestQuery");
const NativeTypeToQuestType = ramda_1.default.invertObj(pg_1.types.builtins);
const timestampDataTypes = [
    pg_1.types.builtins.DATE,
    pg_1.types.builtins.TIMESTAMP,
    pg_1.types.builtins.TIMESTAMPTZ,
];
const timestampTypeParser = (val) => moment.utc(val).format(moment.HTML5_FMT.DATETIME_LOCAL_MS);
class QuestDriver extends query_orchestrator_1.BaseDriver {
    constructor(config = {}) {
        super();
        this.pool = new pg_1.Pool({
            max: process.env.CUBEJS_DB_MAX_POOL && parseInt(process.env.CUBEJS_DB_MAX_POOL, 10) || 4,
            idleTimeoutMillis: 30000,
            host: process.env.CUBEJS_DB_HOST,
            database: process.env.CUBEJS_DB_NAME,
            port: process.env.CUBEJS_DB_PORT,
            user: process.env.CUBEJS_DB_USER,
            password: process.env.CUBEJS_DB_PASS,
            ...config
        });
        this.pool.on('error', (err) => {
            console.log(`Unexpected error on idle client: ${err.stack || err}`);
        });
        this.config = {
            ...this.getInitialConfiguration(),
            ...config,
        };
    }
    static dialectClass() {
        return QuestQuery_1.QuestQuery;
    }
    getInitialConfiguration() {
        return {
            readOnly: true,
        };
    }
    async testConnection() {
        await this.pool.query('SELECT $1 AS number', ['1']);
    }
    mapFields(fields) {
        return fields.map((f) => {
            let questType;
            if (f.dataTypeID in NativeTypeToQuestType) {
                questType = NativeTypeToQuestType[f.dataTypeID].toLowerCase();
            }
            if (!questType) {
                throw new Error(`Unable to detect type for field "${f.name}" with dataTypeID: ${f.dataTypeID}`);
            }
            return ({
                name: f.name,
                type: this.toGenericType(questType)
            });
        });
    }
    async query(query, values, _options) {
        const result = await this.queryResponse(query, values);
        return result.rows;
    }
    async queryResponse(query, values) {
        const conn = await this.pool.connect();
        try {
            const res = await conn.query({
                text: query,
                values: values || [],
                types: {
                    getTypeParser: this.getTypeParser,
                },
            });
            return res;
        }
        finally {
            conn.release();
        }
    }
    getTypeParser(dataTypeID, format) {
        const isTimestamp = timestampDataTypes.includes(dataTypeID);
        if (isTimestamp) {
            return timestampTypeParser;
        }
        const parser = pg_1.types.getTypeParser(dataTypeID, format);
        return (val) => parser(val);
    }
    async downloadQueryResults(query, values, _options) {
        const res = await this.queryResponse(query, values);
        return {
            rows: res.rows,
            types: this.mapFields(res.fields),
        };
    }
    async createSchemaIfNotExists(_schemaName) {
        // no-op as there are no schemas in QuestDB
    }
    async tablesSchema() {
        const tables = await this.getTablesQuery('');
        // QuestDB doesn't have a notion of schema/logical database while the driver
        // has to return a `{ 'schema_name': { 'table1': {...} } }` object. So, we use
        // empty schema name ('') as a workaround to avoid the schema prefix
        // ('schema_name.') being used for table names in the generated queries.
        const metadata = { '': {} };
        // eslint-disable-next-line camelcase
        await Promise.all(tables.map(async ({ table_name: tableName }) => {
            if (tableName === undefined) {
                return;
            }
            const columns = await this.tableColumnTypes(tableName);
            metadata[''][tableName] = columns;
        }));
        return metadata;
    }
    // eslint-disable-next-line camelcase
    async getTablesQuery(_schemaName) {
        const response = await this.query('SHOW TABLES', []);
        return response.map((row) => ({
            table_name: row.table,
        }));
    }
    async tableColumnTypes(table) {
        const response = await this.query(`SHOW COLUMNS FROM ${table}`, []);
        return response.map((row) => ({ name: row.column, type: this.toGenericType(row.type) }));
    }
    async uploadTableWithIndexes(table, columns, tableData, indexesSql) {
        if (!tableData.rows) {
            throw new Error(`${this.constructor} driver supports only rows upload`);
        }
        await this.createTable(table, columns);
        try {
            for (let i = 0; i < tableData.rows.length; i++) {
                await this.query(`INSERT INTO ${table}
        (${columns.map(c => this.quoteIdentifier(c.name)).join(', ')})
        VALUES (${columns.map((c, paramIndex) => this.param(paramIndex)).join(', ')})`, columns.map(c => this.toColumnValue(tableData.rows[i][c.name], c.type)));
            }
            // Make sure to commit the data to make it visible for later queries.
            await this.query('COMMIT', []);
            for (let i = 0; i < indexesSql.length; i++) {
                const [query, params] = indexesSql[i].sql;
                await this.query(query, params);
            }
        }
        catch (e) {
            await this.dropTable(table);
            throw e;
        }
    }
    readOnly() {
        return !!this.config.readOnly;
    }
    release() {
        return this.pool.end();
    }
    param(paramIndex) {
        return `$${paramIndex + 1}`;
    }
}
exports.QuestDriver = QuestDriver;
//# sourceMappingURL=QuestDriver.js.map